/*!
************************************************************************
* \file adaptive_quantization.h
*
* \brief
*    adaptive quantization matrix selection routines
*
* \author
*  Akiyuki Tanizawa       <akiyuki.tanizawa@toshiba.co.jp>           \n
*  Takeshi Chujoh         <takeshi.chujoh@toshiba.co.jp>             \n
*
* This software may be used for research purposes only.
* TOSHIBA would also appreciate that any technical report, conference or
* journal paper which uses the software includes one of the following references:
*    - A. Tanizawa, T. Chujoh, "Adaptive Quantization Matrix Selection",
*                              VCEG contribution D.266, Geneva, November 2006.
*    - A. Tanizawa, T. Chujoh, "Simulation results of AQMS on KTA software",
*                              VCEG contribution VCEG-AC07, Klagenfurt, July 2006.
*    - A. Tanizawa, T. Chujoh, "Adaptive Quantization Matrix Selection on KTA software",
*                              VCEG contribution VCEG-AD06, Hangzhou, Oct. 2006.
*    - A. Tanizawa, T. Chujoh, "Simulation results of AQMS on KTA software version 1.3",
*                              VCEG contribution VCEG-AF08, San Jose, April 2007.
*    - A. Tanizawa, T. Chujoh, T. Yamakage, 
*                              "Encoding complexity reduction of AQMS",
*                              VCEG contribution C.403, Geneve, April 2008.
************************************************************************
*/

/*
* CONFIGURATION:
*
* Configuration of the .Cfg file:
* In section KTA STUFF:
*
* // AQMS command:
* UseAdaptiveQuantMatrix    = 0 (0:disable, 1:enable, 2:fast AQMS (optimal))
* // Adaptive QP command:
* UseAdaptiveQuantParam    = 0 (0:disable, 1:enable)
*
* NOTE:
* Both AQMS and Adaptive QP can perform on VCEG-AA10 conditions.
*    - TK. Tan, G. Sullivan, T. Wedi, "Recommended Simulation Common Conditions for Coding Efficiency Experiments",
*                              VCEG contribution VCEG-AA10, Nice, Oct. 2005.
*    - TK. Tan, G. Sullivan, T. Wedi, "Recommended Simulation Common Conditions for Coding Efficiency Experiments",
*                              VCEG contribution VCEG-AE10r1, Marrakech, Jan. 2007.
********************
* KNOWN LIMITATIONS
********************
* Not compatible with:                
*  - interlace coding
*  - RdOpt != 1   
*  - RateControl
*/

#ifndef _ADAPTIVE_QUANTIZATION_H_
#define _ADAPTIVE_QUANTIZATION_H_

#include "global.h"
#include "mbuffer.h"

#if (defined(ADAPTIVE_QUANTIZATION)||defined(ADAPTIVE_QP))
void PreSetBestCodingParameters(int best_mode, Macroblock* currMB, RD_PARAMS enc_mb, short bslice, short *inter_skip, int best_ctr16x16);
void SubmacroblockCalcCost(RD_PARAMS, RD_8x8DATA *, Macroblock *,int ***, int *, short, int, int *, int *, int *, int);
void ComputeQuantizationRDcost(int mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcost, double *min_rate, int i16mode, short bslice, short *inter_skip,
                               short best_ctr16x16, short best_cbp, short *best_mode, short *best_matrix, short *best_qp);

void ComputeBestQuantizationRDCost( int best_mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcost, double *min_rate, int i16mode, short bslice, int rd_qp,
                                   short *best_bi_pred_me, short *inter_skip, short best_ctr16x16, short best_cbp, short *best_matidx, short *best_qp);
void InitAdaptiveQuantizationParam (void);
void CalculateLambdaForMatrix(RD_PARAMS   *enc_mb);
#endif

#ifdef ADAPTIVE_QUANTIZATION
//------ for image.c
typedef struct
{
  int default_AQMS_flag;
  int default_Coeff_flag;
  int default_QMOpt_flag;
  int default_me_modeflag;
  int default_bipred_me_flag;
  int default_BiPredMESearchRange;
  int default_ChromaIntraDisable;
  int default_rdopt;
  int default_SelectiveIntraEnable;
  int default_EarlySkipEnable;
  
#ifdef ADAPTIVE_FD_SD_CODING
  int default_APEC_in_FD_and_SD;
#endif
#ifdef MV_COMPETITION
  int default_MVCompetition;
#endif
} sDefaultInputParam_t;

int SliceMatrixParam0[5];
int SliceMatrixParam1[5];

extern void ReGenerateParameterSets (int best_pic_id);
int rdPictureCodingAQMS(int AQMS_Skip_flag, sDefaultInputParam_t *defInputParam, int *mincost_flg, int rd_qp);
int rdPictureCodingForModelingMatrixDesign(int rd_qp);
void PictureCodingDecisionForMatrixDesign(sModelingMatrixType_t *ModelingMatrixType, int param, int *Rate1, int *Dist1, int *Rate2, int *Dist2, double *lambda, int rd_qp);
void CalculateInitinalModelingParamters(int param_idx, int *ParamA, int *ParamB);
void SetActiveMatrixFromParam(int Idx, sModelingMatrixType_t *ModelingMatrixType);
void PictureCodingDecisionForMatrix(sModelingMatrixType_t *ModelingMatrixType, int *Rate, int *Dist, double *lambda, int rd_qp);

void SaveInputParam(int AQMS_Skip_flag, sDefaultInputParam_t *defInputParam);
void LoadInputParam(sDefaultInputParam_t *defInputParam);
void SetFastEncoding(sDefaultInputParam_t *defInputParam);
void SetFastEncodingForAQMS(sDefaultInputParam_t *defInputParam);

void set_selective_matrix(int *cur_slice_bit_pos);
void compute_R_D_data(int cur_slice_bit_pos);

int  CalculateBestMatrixParam(int param_idx, int S1, int S2, int R1, int R2, int D1, int D2, double lambda, int rd_qp);
void SetModelingMatrixParam(sModelingMatrixType_t *ModelingMatrixType, int type, int gradient, int DCgain);
void ResetFrameMemory();
int  CheckModelingParam(int param_idx, double EstParam);
void VerifyModelingParam(sModelingMatrixType_t *ModelingMatrixType);
int  CheckSameModelingParam();
int  ChooseBestPicture(int rd_qp);
void ResetModelingParam (void);

int best_rd_pass;
void StoreMotionVectorForAQMS();
void RestoreMotionVectorForAQMS();
StorablePicture *enc_frame_picture_aqms;

//------ for mode_decision.c
void DecideBestMatrixForMB(int mode,  Macroblock *currMB,  RD_PARAMS enc_mb,  double *min_rdcost, double *min_rate, int i16mode,  short bslice, 
                           short *inter_skip, short best_ctr16x16, short *best_mode, short *best_matrix);
void DecideBestRdoForMB(int mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcost, double *min_rate, int i16mode, short bslice, short *inter_skip,
                        short best_ctr16x16, short *best_rdo);
void DecideBestCoeffForMB(int mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcost, double *min_rate, int i16mode, short bslice, short *inter_skip,
                          short best_ctr16x16, short *best_rdo);
void ComputeCoeffRDCost(int mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcost, double *min_rate, int i16mode, short bslice, short *inter_skip,
                        short best_ctr16x16, short best_cbp);

//------ for q_matrix.c
void InitScalingListParam (void);
void SetAutoScalingListData(int InternalFlag);
void CreateScalingList4x4(int pT, int pA, int pB, int InternalFlag);
void CreateScalingList8x8(int pT, int pA, int pB, int InternalFlag);

#ifdef MV_COMPETITION
void ReEncodeMVCompetition(int rd_pass);
#endif
#endif

#ifdef ADAPTIVE_QP
unsigned int useQPvalueFlag;
void DecideBestQPForMB(int mode, Macroblock *currMB, RD_PARAMS enc_mb, double *min_rdcost, double *min_rate, int i16mode, short bslice,
                       short *inter_skip, short best_ctr16x16, short *best_mode, short *best_qp);
#endif

#endif //_ADAPTIVE_QUANTIZATION_H_

